/*
    ext2online.c -- resizes a mounted ext2 fs
    Copyright (C) 1999 Andreas Dilger <adilger@enel.ucalgary.ca>
    Copyright (C) 1999 Lennert Buytenhek <buytenh@dsv.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  This program is to be used in cooperation with Andreas Dilger's
  ext2-online resize patch for the linux kernel 2.0, 2.2, or 2.3.
*/

static const char _ext2online_c[] = "$Id: ext2online.c,v 1.8 1999/10/05 19:15:57 buytenh Exp $";

#include "config.h"

#ifndef HAVE_LINUX_EXT2_FS_H
#error This program only works on linux
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <mntent.h>
#include <sys/mount.h> /* mount.h may define these next two, but if not... */
#if !defined(MS_MGC_VAL) || !defined(MS_REMOUNT)
#include <linux/fs.h>
#endif
#include "ext2.h"

#ifndef MNTTYPE_EXT2
#define MNTTYPE_EXT2 "ext2"
#endif

char			*progname;
int			 opt_debug = 0;
int			 opt_verbose = 0;
int			 opt_ignore = 0;

static void print_version(FILE *outfile)
{
	fprintf(outfile, "ext2online v%s - 1999/10/02 for EXT2FS %s\n",
		VERSION, EXT2FS_VERSION);
}

static void usage(FILE *outfile)
{
	fprintf(outfile,
		"usage: %s [-d] [-i] [-v] device [new_size[bkmgt]]\n"
		"\t-d, --debug    : turn debug info on\n"
		"\t-i, --ignore   : ignore device size check\n"
		"\t-v, --verbose  : be verbose\n"
		"\tnew_size is in ext2 filesystem blocks (1k, 2k, or 4k),\n"
		"\tdisk Blocks (512 byte), Kilo-, Mega-, Giga-, or Terabytes\n",
		progname);
	exit(1);
}

void parse_args(int *argc, char **argv[])
{
	static struct option long_options[] = {
		{"debug", 0, 0, 'd'},
		{"ignore", 0, 0, 'i'},
		{"verbose", 0, 0, 'v'},
		{NULL, 0, 0, 0}};
	int c;
	int dummy;

	while ((c = getopt_long(*argc, *argv, "div", long_options, &dummy)) != -1)
	{
		switch (c)
		{
		case 'd': opt_debug = 1; break;
		case 'i': opt_ignore = 1; break;
		case 'v': opt_verbose = 1; break;
		default:  fprintf(stderr, "%s: unknown option %c\n",
				  progname, c);
			  usage(stderr);
		}
	}

	*argc -= optind;
	*argv += optind;
}

/* Find the mountpoint or device from the string in *dev */
static void find_mount(char *filename, char **dev, char **dir, char **type,
		       char **opts)
{
	FILE		*tabfile;
	struct mntent	*ent;

	tabfile = setmntent(filename, "r");

	while  ((ent = getmntent(tabfile)) != NULL)
	{
		if (strcmp(*dev, ent->mnt_fsname) == 0)
		{
			if (strcmp(ent->mnt_type, MNTTYPE_EXT2) != 0)
				break;
			*dir = (char *)malloc(strlen(ent->mnt_dir) + 1);
			if (*dir == NULL)
				*dir = ent->mnt_dir;
			else
				strcpy(*dir, ent->mnt_dir);
			break;
		}
		else if (strcmp(*dev, ent->mnt_dir) == 0)
		{
			if (strcmp(ent->mnt_type, MNTTYPE_EXT2) != 0)
				break;
			*dir = *dev;
			*dev = (char *)malloc(strlen(ent->mnt_fsname) + 1);
			if (*dev == NULL)
				*dev = ent->mnt_dir;
			else
				strcpy(*dev, ent->mnt_fsname);
			break;
		}
	}
	if (*dir != NULL && ent != NULL)
	{
		*type = (char *)malloc(strlen(ent->mnt_type) + 1);
		if (*type == NULL)
			*type = ent->mnt_type;
		else
			strcpy(*type, ent->mnt_type);
		/* Leave room at end for ",resize=NNNNNNNNNN:NNNN" parameter */
		*opts = (char *)malloc(strlen(ent->mnt_opts) + 25);
		if (*opts == NULL)
		{
			fprintf(stderr,
			"%s: no memory for %s mount options, resize aborted\n",
				progname, *dev);
			*dir = NULL;
		}
		else
		{
			char *index, *index2;

			if (hasmntopt(ent, MNTOPT_RO))
			{
				fprintf(stderr,
				"%s: %s is mounted read-only, resize aborted\n",
					progname, *dev);
				*dir = NULL;
			}

			/* Skip the "rw" and "nosuid" options, keep others */
			if ((index = hasmntopt(ent, MNTOPT_NOSUID)) != NULL)
			{
				if ((index2 = strchr(index, ',')) == NULL)
					index[0] = '\0';
				else
					while ((*index++ = *++index2) != '\0');
			}

			if ((index = hasmntopt(ent, MNTOPT_RW)) != NULL)
			{
				if ((index2 = strchr(index, ',')) == NULL)
					*index = '\0';
				else
					while ((*index++ = *++index2) != '\0');
			}

			if (ent->mnt_opts[0] == '\0')
				*opts[0] = '\0';
			else
				sprintf(*opts, "%s,", ent->mnt_opts);
		}
	}
	endmntent(tabfile);
}


/* Check that all of the needed group descriptor blocks are reserved */
int ext2_check_newgdblocks(struct ext2_fs *fs, struct ext2_inode *inode,
			blk_t newgdblocks)
{
	int	group;
	blk_t	block;
	blk_t	gdb;

	for (group = 0, gdb = fs->sb.s_first_data_block + 1;
	     group < fs->numgroups; group++, gdb += fs->sb.s_blocks_per_group)
	for (block = fs->gdblocks; block < newgdblocks; block++)
	{
		if (fs->opt_debug)
			printf("checking for group block %d in Bond\n", gdb);

		if (ext2_is_group_sparse(fs, group) &&
		    ext2_do_inode(fs, inode, gdb + block, EXT2_ACTION_FIND) < 0)
		{
			fprintf(stderr, "%s: group %d, block %u not reserved\n",
				progname, group, gdb + block);
			return 0;
		}
	}
	return 1;
}

#define mark_block(bh, offset) (bh)->data[(offset)>>3] |= _bitmap[(offset)&7]

/* Create group descriptors, bitmaps, inodes for each new group */
blk_t ext2_make_group(struct ext2_fs *fs, struct ext2_inode *inode, int group,
		      blk_t blocks, blk_t newgdblocks, blk_t skip,
		      blk_t resgdblocks)
{
	struct ext2_buffer_head	*bh;
	struct ext2_group_desc	*gdp;
	blk_t			 start;
	blk_t			 bboffset, iboffset, itoffset;
	int			 max_inodes;
	int			 i;

	if (fs->opt_verbose)
		printf("%s: creating group %d with %u blocks "
		       "(skip = %u, res = %u, newgd = %u)\n", progname,
		       group, blocks, skip, resgdblocks, newgdblocks);

	if (blocks < fs->adminblocks + 50)
	{
		/* This would only happen on the last group */
		fprintf(stderr, "%s: not enough blocks (%u) for group %d\n",
			progname, blocks, group);
		return 0;
	}

	start = fs->sb.s_first_data_block + group * fs->sb.s_blocks_per_group;

	itoffset = fs->itoffset;
	if (itoffset + fs->inodeblocks > blocks)
	{
		printf("%s: resetting metadata offset for group %d\n",
			progname, group);
		itoffset = 3 + ext2_is_group_sparse(fs, group) ?
			   newgdblocks + resgdblocks : 0;
	}

	/* Create the primary copy of the group descriptor table */
	bh = ext2_bread(fs, fs->sb.s_first_data_block +
			howmany((group + 1) * sizeof(struct ext2_group_desc),
				fs->blocksize));
	bh->dirty = 1;
	gdp = (struct ext2_group_desc *)bh->data +
	      (group % sizeof(struct ext2_group_desc));

	if (skip)
	{
		blk_t next = itoffset + fs->inodeblocks;
		blk_t newoff = next + skip * group;

		skip = newoff >= blocks ? next : newoff;
	}

	if (ext2_is_group_sparse(fs, group))
	{
		bboffset = skip ? skip : itoffset - 2;
		gdp->bg_free_blocks_count = 0;
	}
	else
	{
		bboffset = skip ? skip : 1 + newgdblocks + resgdblocks;
		gdp->bg_free_blocks_count = 1 + newgdblocks + resgdblocks;
	}
	iboffset = bboffset + 1;

	gdp->bg_block_bitmap = start + bboffset;
	gdp->bg_inode_bitmap = start + iboffset;
	gdp->bg_inode_table  = start + itoffset;
	gdp->bg_free_blocks_count += blocks - fs->adminblocks;
	gdp->bg_free_inodes_count = fs->sb.s_inodes_per_group;
	gdp->bg_used_dirs_count = 0;
	gdp->bg_pad = 0;
	gdp->bg_reserved[0] = 0;
	gdp->bg_reserved[1] = 0;
	gdp->bg_reserved[2] = 0;
	/* Update the in-memory copy of the group descriptor as well */
	memcpy(&fs->gd[group], gdp, sizeof(struct ext2_group_desc));
	ext2_brelse(bh, 0);

	/* Clear metadata blocks to show all blocks/inodes unused */
	if (fs->opt_debug)
	{
		printf("clear/mark block bitmap 0x%04x used\n",
			start + bboffset);
		printf("clear/mark inode bitmap 0x%04x used\n",
			start + iboffset);
		printf("clear/mark inode blocks 0x%04x-0x%04x used\n",
			start + itoffset,
			start + itoffset + fs->inodeblocks - 1);
	}

	ext2_zero_blocks(fs, start + bboffset, 1);
	ext2_zero_blocks(fs, start + iboffset, 1);
	ext2_zero_blocks(fs, start + itoffset, fs->inodeblocks);

	/* Mark in-use metadata blocks */
	bh = ext2_bread(fs, start + bboffset);
	bh->dirty = 1;
	mark_block(bh, bboffset);
	mark_block(bh, iboffset);
	for (i = itoffset; i < itoffset + fs->inodeblocks; i++)
		mark_block(bh, i);

	/* Mark metadata blocks used in block bitmap */
	if (ext2_is_group_sparse(fs, group))
	{
		if (fs->opt_debug)
		{
			printf("mark superblock 0x%04x used\n", start);
			printf("mark group desc. 0x%04x-0x%04x used\n",
				start + 1, start + newgdblocks);
			if (resgdblocks)
				printf("mark reserved descriptors "
					"0x%04x-0x%04x used\n",
					start + newgdblocks + 1,
					start + newgdblocks + resgdblocks);
		}

		mark_block(bh, 0);		/* superblock */
		for (i = 1; i <= newgdblocks; i++)
			mark_block(bh, i);	/* group descriptor blocks */
		for (; i <= newgdblocks + resgdblocks; i++)
		{
			mark_block(bh, i);	/* reserved GDT blocks */
			ext2_do_inode(fs, inode, start + i, EXT2_ACTION_ADD);
		}
	}

	/* Mark unavailable blocks at end of group used */
	if (fs->opt_debug && blocks != fs->blocksize * 8)
		printf("mark end blocks 0x%04x-0x%04x used\n",
			start + blocks, start + fs->blocksize * 8 - 1);

	for (i = blocks; i < fs->blocksize * 8; i++)
		mark_block(bh, i);
	ext2_brelse(bh, 0);

	/* Mark unavailable inodes at end of table used */
	max_inodes = fs->blocksize * 8;
	if (max_inodes != fs->sb.s_inodes_per_group)
	{
		if (fs->opt_debug)
			printf("mark unavailable end inodes used\n");

		bh = ext2_bread(fs, iboffset + start);
		bh->dirty = 1;
		for (i = fs->sb.s_inodes_per_group; i < max_inodes; i++)
			bh->data[i>>3] |= _bitmap[i&7];
		ext2_brelse(bh, 0);
	}

	return blocks;
}

/* Create new groups if necessary to handle new filesystem size */
blk_t ext2_online_primary(struct ext2_fs *fs, struct ext2_inode *inode,
			  blk_t newsize, int newgroups, blk_t newgdblocks,
			  blk_t resgdblocks)
{
	int			 group;
	int			 skip = 0;
	blk_t			 realsize;

	if (fs->opt_verbose)
		printf("%s: resize to %u blocks\n", progname, newsize);

	if (newsize <= fs->sb.s_blocks_count)
		return newsize;

	if (fs->opt_debug)
	{
		printf("%d old groups, %d blocks\n", fs->numgroups,
			fs->gdblocks);
		printf("%d new groups, %d blocks\n", newgroups,
			newgdblocks);
	}

	if (newgdblocks > fs->gdblocks)	/* We need to use more GDT blocks */
	{
#if 1
		if (!ext2_check_newgdblocks(fs, inode, newgdblocks))
			return 0;
#else
		printf("Not checking Bond for blocks!!!!\n");
#endif
	}

#if 0
	if (fs->numgroups > 2)	/* Try to keep RAID spacings */
	{
		skip = fs->gd[2].bg_block_bitmap - fs->gd[1].bg_block_bitmap;
		if (skip != fs->gd[2].bg_inode_bitmap -
			    fs->gd[1].bg_inode_bitmap)
			skip = 0;
	}
#endif

	/* Fill last existing group */
	realsize = min(fs->numgroups * fs->sb.s_blocks_per_group +
			fs->sb.s_first_data_block, newsize);
	if (fs->opt_debug && realsize > fs->sb.s_blocks_count)
		printf("Filling last group to %u blocks\n", realsize);

	fs->adminblocks += newgdblocks + resgdblocks - fs->gdblocks;
	/* Add new groups */
	for (group = fs->numgroups; group < newgroups; group++)
	{
		blk_t blocks;

		blocks = min(fs->sb.s_blocks_per_group, newsize - realsize);
		realsize += ext2_make_group(fs, inode, group, blocks,
					    newgdblocks, skip, resgdblocks);
	}

	return realsize;
}

/* Copy primary superblock, group descriptors to all "sparse" groups after
 * we are sure that the resize has worked
 */
void ext2_online_finish(struct ext2_fs *fs, struct ext2_inode *inode,
			blk_t newblocks, int newgroups, blk_t newgdblocks,
			blk_t resgdblocks)
{
	int			 group;
	blk_t			 block;
	blk_t			 gdb;

	/* Remove blocks as required from Bond inode */
	for (group = 0, gdb = fs->sb.s_first_data_block + 1;
	     group < fs->numgroups; group++, gdb += fs->sb.s_blocks_per_group)
		if (ext2_is_group_sparse(fs, group))
			for (block = fs->gdblocks; block < newgdblocks; block++)
			{
				if (fs->opt_debug)
					printf("Checking for group block %d "
						"in Bond\n", gdb);
				ext2_do_inode(fs, inode, gdb + block,
					      EXT2_ACTION_DELETE);
			}

	/* Re-read kernel-updated superblock from disk, and fix up the rest
	 * of the fs structure so that the backup superblock copies are OK.
	 */
	fs->devhandle->ops->read(fs->devhandle->cookie, &fs->sb, 1, 1);
	fs->gdblocks = newgdblocks;
	fs->numgroups = newgroups;
	fs->metadirty = EXT2_META_BACKUP;
}

blk_t get_reserved(struct ext2_fs *fs, struct ext2_inode *inode,
		   blk_t newgdblocks)
{
	blk_t	resgdblocks;

#if 0
	/* How many blocks to reserve for future GDT expansion.  This assumes
	 * Bond has the right number of blocks assigned.  It may not be
	 * reliable unless e2fsck starts fixing Bond for us.
	 */
	resgdblocks = (inode->i_blocks - inode->i_block[EXT2_IND_BLOCK] ? 1:0 -
		  inode->i_block[EXT2_DIND_BLOCK] ? 1 : 0 -
		  inode->i_block[EXT2_TIND_BLOCK] ? 1 : 0) / 2 / fs->numgroups;

#else
	/* FIXME: this only works for normal, non-sparse, non-striped FS
	 *        that has been modified by ext2prepare only.
	 */
	resgdblocks = fs->itoffset - newgdblocks - 3;
	resgdblocks = resgdblocks > 100 ? 0 : resgdblocks; /* Bogus val test */
	if (fs->opt_debug)
		printf("using %u reserved GDT blocks\n", resgdblocks);
#endif
	return resgdblocks;
}

int main(int argc, char *argv[])
{
	char			*dev, *type, *opts, *dir = NULL;
	struct ext2_dev_handle	*handle;
	struct ext2_fs		*fs;
	struct ext2_inode	 inode;
	int			 newgroups;
	blk_t			 newsize = 0, devsize = 0, realsize = 0;
	blk_t			 newgdblocks, resgdblocks;
	char			 mod = '\0';

	progname = strrchr(argv[0], '/') ? strrchr(argv[0], '/') + 1 : argv[0];
	print_version(stdout);
	parse_args(&argc, &argv);

	if (argc < 1 || argc > 2)
		usage(stderr);

	dev = argv[0];

	find_mount(MOUNTED, &dev, &dir, &type, &opts);
	if (dir == NULL)
	{
		fprintf(stderr, "%s: can't find %s, is it mounted?\n",
			progname, dev);
		usage(stderr);
	}

	handle = ext2_make_dev_handle_from_file(dev);
	if (handle == NULL)
	{
		fprintf(stderr, "%s: can't open %s\n", dev, progname);
		return 1;
	}

	if (argc == 2)
	{
		sscanf(argv[1], "%i%c", &newsize, &mod);

		switch (mod)	/* Order of these options is important!! */
		{
			case 't':
			case 'T': newsize *= 1024; /* no break */
			case 'g':
			case 'G': newsize *= 1024; /* no break */
			case 'm':
			case 'M': newsize *= 1024; /* no break */
			case 'k': /* Need to open fs to find block size */
			case 'K':
			case  0 : break;
			case 'b':
			case 'B': newsize /= 2; break;
			default: usage(stderr);
		}

		newgroups = howmany(newsize, 8192); /* worst case 1k blocks */

		if (ext2_max_groups < newgroups)
			ext2_max_groups = newgroups;
	}

	if ((fs = ext2_open(handle, EXT2_VALID_FS)) == NULL)
	{
		fprintf(stderr, "%s: can't open %s\n", dev, progname);
		return 1;
	}
	fs->online = 1;
	fs->opt_debug = opt_debug;
	fs->opt_verbose = opt_verbose;

	if (mod != '\0')
		newsize /= fs->blocksize / 1024;

	devsize = fs->devhandle->ops->get_size(fs->devhandle->cookie);

	if (argc == 1)	/* Resize to the full size if none is given */
		newsize = devsize;
	else if (newsize > devsize && !opt_ignore)
	{
		/* We could put a system() call here to lvextend with the
		 * appropriate parameters to grow the LV before the resize.
		 */
		fprintf(stderr, "%s: %s has %u blocks not %u, resize aborted\n",
			progname, dev, devsize, newsize);
		return 1;
	}
	else if (newsize <= fs->sb.s_blocks_count)
	{
		/* Not currently, but it would be possible if the groups at
		 * the end of the FS are not in use at all (no inode, blocks)
		 * We could even do a system() call to lvreduce after the
		 * mount() call returned, when we can get a status code back...
		 */
		fprintf(stderr, "%s: %s has %u blocks cannot shrink to %u\n",
			progname, dev, fs->sb.s_blocks_count, newsize);
		return 1;
	}

	if (newsize >= 1<<28 && fs->blocksize == 1024)
	{
		fprintf(stderr, "%s: can't resize 1k block filesystem over "
			"256GB.\n\tToo much space is wasted.\n", progname);
		return 1;
	}

	newgroups = howmany(newsize - fs->sb.s_first_data_block,
			    fs->sb.s_blocks_per_group);
	newgdblocks = howmany(newgroups * sizeof(struct ext2_group_desc),
			      fs->blocksize);
	/* Bring in Mr. Bond */
	ext2_read_inode(fs, EXT2_RESIZE_INO, &inode);

	/* How many blocks to reserve for future GDT expansion */
	resgdblocks = get_reserved(fs, &inode, newgdblocks);

	if ((realsize = ext2_online_primary(fs, &inode, newsize, newgroups,
					    newgdblocks, resgdblocks)) == 0)
	{
		fprintf(stderr, "%s: can't resize past %u blocks\n",
			progname, fs->gdblocks *  fs->sb.s_blocks_per_group *
			fs->blocksize / sizeof(struct ext2_group_desc));
		return 1;
	}
	else if (realsize == fs->sb.s_blocks_count)
	{
		fprintf(stderr, "%s: unable to resize %s to %u blocks\n",
			progname, dev, newsize);
		return 1;
	}

	printf("%s: resizing to %u blocks\n", progname, realsize);

	ext2_write_inode(fs, EXT2_RESIZE_INO, &inode);
	ext2_sync(fs);

	/* Add the resize option to the end of the other mount options */
	sprintf(opts + strlen(opts), "resize=%u", realsize);
	if (resgdblocks)
		sprintf(opts + strlen(opts), ":%u", resgdblocks);

	if (fs->opt_debug)
	{
		printf("Calling mount() for %s(%s) with %s\n", dev, dir, opts);
#if 0
		printf("Press ENTER to continue:\n");
		mod = getchar();
#endif
	}
	if (mount(dev, dir, type, MS_MGC_VAL | MS_REMOUNT, opts) < 0)
	{
		fprintf(stderr,"%s: resize failed while in kernel\n", progname);
		perror(progname);
		return 1;
	}


	/* Clean up the rest of the filesystem and the Bond inode */
	ext2_online_finish(fs, &inode, realsize, newgroups, newgdblocks,
			   resgdblocks);
	ext2_write_inode(fs, EXT2_RESIZE_INO, &inode);
	ext2_close(fs);

	return 0;
}
